# -*- coding: utf-8 -*-

"""
Python library for asset pricing research.

"""

__title__ = 'pyanomaly'
__version__ = '0.9'
__author__ = 'Chulwoo Han and Jongho Kang'
__copyright__ = '2021 @ Chulwoo Han and Jongho Kang'

from pyanomaly.globals import *
from pyanomaly.fileio import *
